# -*- coding: utf-8 -*-

# FOGLAMP_BEGIN
# See: http://foglamp.readthedocs.io/
# FOGLAMP_END

""" Server functions for plugin module to detect need for large discharge (LD) """


__author__ = "Amandeep Singh Arora"
__copyright__ = "Copyright (c) 2021 Dianomic Systems"
__license__ = "Apache 2.0"
__version__ = "${VERSION}"

import os
import joblib
import datetime

import numpy as np
import wpt_features as wpt


def adjust_CPT5_columns(X):
    """ adjust_CPT5_columns - adjust CPT5 column values in lowest frequency node of WPD
        Arguments:
            X - input feature data
        Returns:
            processed feature data
    """
    X[:,320*0+16] = np.log(X[:,320*0+16])
    X[:,320*1+16] = np.log(X[:,320*1+16])
    X[:,320*2+16] = np.log(X[:,320*2+16])
    X[:,320*3+16] = np.log(X[:,320*3+16])

    return X


def check_anomaly(self, handle):
    """ check_anomaly - run model inference to look for anomaly in time series data
        Arguments:
            handle - configuration options
        Returns:
            y_pred - array of predictive values; all zeros indicate a perfect match for expected data
                     all ones indicate a perfect match for anomalous data
            RE_list - for each entry above, the reconstruction error (RE) to RE threshold ratio values
    """

    t1 = datetime.datetime.now()

    if (handle['modelType']['value'] in ['Neural network', 'Autoencoder']) and self.interpreter is None:
        # sanity check
        self.error("Neural net interpreter not present")
        raise ValueError

    try:
        interval_rows = int(self.frame_size*1)

        # extract features from time series data using wavelet packet decomposition
        features = wpt.parse_vibration_data(self.frame, self.sampling_rate,  interval_rows, 4)
        features = np.array(features)

        # adjust CPT5 column values in lowest frequency band in WPD
        features2 = adjust_CPT5_columns(features)

        # normalize 
        features3 = self.scaler.transform(features2)

        # do principal component analysis to extract the most important features
        features4 = self.pca.transform(features3)

        del features, features2, features3

        # -- multiple ml choices --
        # either:
        # . use xgboost to create a decision tree of features that match or don't match "healthy" centrifuge state
        # . use a DNN to classify the input as normal / not normal based on wpd training
        # . use an autoencoder DNN to decide if the output is too far from the input to be a normal mapping

        modelType = handle['modelType']['value']
        RE_list = []

        if modelType == 'Xgboost':
            y_pred = self.xgb_clf.predict(features4)
            # threshold = 0.9
            self.y_pred = np.append(self.y_pred, y_pred)
        elif modelType == 'Autoencoder':
            threshold = self.model_data['RE_threshold_uint8_quant'] if self.tpu else self.model_data['RE_threshold']
            y_pred,RE_list = self.ae_prediction(features4, self.tpu, self.interpreter, threshold)
            self.y_pred = np.append(self.y_pred, y_pred)
        else:
            self.logger.debug("calling nn")
            
            y_pred = self.nn_prediction(features4, self.tpu, self.interpreter)
            # threshold = 0.8
            y_pred = y_pred.round()  # binarize probability to one of the 2 classes

            self.y_pred = np.append(self.y_pred, y_pred)

        if self.y_pred.shape[0] > self.max_prediction:
            to_del = self.y_pred.shape[0]-self.max_prediction
            self.y_pred = np.delete(self.y_pred, slice(0,to_del), 0)

        self.logger.info("y_pred={} -> {} ({})".format(y_pred, self.y_pred, np.count_nonzero(self.y_pred)))
        self.logger.info("------ self.y_pred: sum={}, length={}, ratio={}, num_frames_processed={}, num_input_rows={}, so rv={}".
                         format(self.y_pred.sum(), len(self.y_pred),
                                (self.y_pred.sum() / len(self.y_pred)).round(2),
                                self.num_rows_processed // self.frame_size,
                                self.num_input_rows,
                                np.count_nonzero(self.y_pred) >= self.min_nonzero_anomaly))

        y_pred = 1 if (np.count_nonzero(self.y_pred) >= self.min_nonzero_anomaly) else 0

    except Exception as ex:
        self.logger.exception('Error in evaluating ADM LD plugin {}'.format(str(ex)))
        raise

    t2 = datetime.datetime.now()
    delta = t2 - t1
    self.logger.debug("check_anomaly took {} usec, returning y_pred={}".format(int(delta.total_seconds() * 1000 * 1000), y_pred))

    return y_pred, RE_list   # y_pred: 0 means LD not required, 1 means LD required


def ae_prediction(self, features4, tpu, interpreter, threshold):
    """ get prediction on features list using auto encoder
        Args:
            features4: processed WPD statistical features
            interpreter: tensorflow interpreter
            tpu: bool, use tpu or not
            threshold: Maximum reconstruction error for normal centrifuge operation
        Returns:
            Returns prediction for each feature row in the input
    """

    features4 = features4.astype('float32')

    # output = interpreter.tensor(interpreter.get_output_details()[0]["index"])
    # self.logger.info("interpreter.get_input_details()={}".format(interpreter.get_input_details()))
    # self.logger.info("interpreter.get_output_details()={}".format(interpreter.get_output_details()))

    if tpu:
        # for TPU normalize to a zero point (then reverse after creating prediction)
        v = interpreter.get_input_details()[0]['quantization']
        scale = v[0]
        zero_point = v[1]
        features4 = np.round(features4 / scale + zero_point).astype('uint8')

    t1 = datetime.datetime.now()

    y_pred = []
    mse_list = []
    for r in range(features4.shape[0]):
        feat = np.expand_dims(features4[r], axis=0)
        # for each of the input feature row, create an output prediction
        interpreter.set_tensor(interpreter.get_input_details()[0]['index'], feat)
        interpreter.invoke()
        prediction = interpreter.get_tensor(interpreter.get_output_details()[0]['index'])
        # self.logger.info("interpreter.get_output_details()={}".format(interpreter.get_output_details()))

        if tpu:
            # reverse the normalization
            v = interpreter.get_output_details()[0]['quantization']
            scale = v[0]
            zero_point = v[1]
            prediction = (prediction.astype('float32') - zero_point) * scale

        prediction = np.array(prediction)
        mse = np.mean(np.power(feat - prediction, 2), axis=1)
        self.logger.info("r={}, mse={}", r, mse)
        mse_list.append(mse[0])
        y_pred.append(1 if mse >= threshold else 0)

    # return the prediction -- a majority of outputs need to signal in order to be classified as anomalous
    y_pred = np.array(y_pred)
    self.logger.debug("threshold={}, mse_list={}".format(threshold, [ "{:.8f}".format(mse) for mse in mse_list] ))
    t2 = datetime.datetime.now()
    delta = t2 - t1
    self.logger.debug("Inference done in {} usec".format(int(delta.total_seconds() * 1000 * 1000)))

    RE_list = ["{:.4f}".format(k/threshold) for k in mse_list]
    return y_pred, RE_list


def ml_plugin_init(self, handle):
    """ plugin_init - ml-function specific values that
        modifies handle dict containing configuration values
        sets ModelServer state related to config items
        Args:
            handle: configuration options
        Returns:
            modified handle
    """

    # accumulate 0/1 from the past n frames in an array; anomaly is when m/n are 1
    self.y_pred = np.array(0)

    self.min_nonzero_anomaly = int(handle['numAnomalies']['value'])
    self.max_prediction = int(handle['numEvaluations']['value'])
    
    # set up the various model files used to do ML/NN/AE analysis
    model_dir = handle['modelDirPath']['value']
    modelType = handle['modelType']['value']

    if modelType in ['Neural network', 'Autoencoder']:
        handle['enableTpu'] = handle['enableTpu']['value']
        abbr = "nn" if (modelType == 'Neural network') else "ae"

        if handle['enableTpu'] == 'true':
            handle['modelPath'] = os.path.join(model_dir, "adm_ld_"+abbr+"_edgetpu.tflite")
        else:
            handle['modelPath'] = os.path.join(model_dir, "adm_ld_"+abbr+"_float32.tflite")
    else:
        handle['modelPath'] = os.path.join(model_dir, "adm_ld_xgb.model")

    handle['modelDataPath'] = os.path.join(model_dir, "ae_model_data.npy" if (modelType == 'Autoencoder') else "model_data.npy")
    handle['scalerFilePath'] = os.path.join(model_dir, "ae_min_max_scaler.dat" if (modelType == 'Autoencoder') else "min_max_scaler.dat")
    handle['pcaFilePath'] = os.path.join(model_dir, "ae_kernel_PCA_64.dat")

    if not os.path.isfile(handle['pcaFilePath']):
        self.logger.info("{} not found, will attempt to create it by combining chunks".format(handle['pcaFilePath']))
        if os.path.isfile(handle['pcaFilePath']+'.aa'):
            self.logger.info("Combining chunks of {} file".format(handle['pcaFilePath']))
            os.system("cat {}.?? > {}".format(handle['pcaFilePath'], handle['pcaFilePath'] ))
        else:
            self.logger.error("Chunks of {} also not found".format(handle['pcaFilePath']))

    for file in ['modelPath', 'modelDataPath', 'pcaFilePath', 'scalerFilePath']:
        if not os.path.isfile(handle[file]):
            raise Exception("Unable to find file: {}".format(handle[file]))
        else:
            self.logger.debug("Found file: {}".format(handle[file]))
            pass

    # load saved data pre-processing objects required for transforming data into features
    self.pca = joblib.load(handle['pcaFilePath'])
    self.scaler = joblib.load(handle['scalerFilePath'])
    self.model_data = np.load(handle['modelDataPath'], allow_pickle=True).item()
    
    if modelType == 'Xgboost':
        self.xgb_clf = joblib.load(handle['modelPath'])

    elif modelType in ['Neural network', 'Autoencoder']:
        self.tpu = handle['enableTpu'] == 'true'

        # create the tensorflow runtime,
        try:
            from tflite_runtime.interpreter import Interpreter

            try:
                from tflite_runtime.interpreter import load_delegate
            except ImportError:
                self.logger.exception("Edge TPU Support is not found in Tensorflow installation ")
            except:
                self.logger.exception("TFlite runtime installation not found")
                raise
            
            t1 = datetime.datetime.now()

            if handle['enableTpu'] == 'true':
                # load the Edge TPU Runtime
                interpreter = Interpreter(model_path=handle['modelPath'],
                                          experimental_delegates=[load_delegate('libedgetpu.so.1.0')])
            else:
                interpreter = Interpreter(model_path=handle['modelPath'])

            interpreter.allocate_tensors()
            self.interpreter = interpreter
                
            t2 = datetime.datetime.now()
            delta = t2 - t1
            self.logger.info("Neural network model loaded in {:.2f} msec".format(delta.total_seconds() * 1000))
        except Exception as e:
            self.interpreter = None
            self.logger.error('Failed to create TFlite interpreter instance: {}{}'.
                          format(e, "   Check whether edge TPU is attached" if handle['enableTpu'] == 'true' else ""))
            raise

    # return modified handle
    return handle


export_service_functions=[check_anomaly, ae_prediction, ml_plugin_init]
